//IMPORTS

import deng.fzip.FZip;
import deng.fzip.FZipFile;

import flash.desktop.NativeApplication;
import flash.events.Event;
import flash.events.ProgressEvent;
import flash.events.ServerSocketConnectEvent;
import flash.events.TimerEvent;
import flash.filesystem.File;
import flash.net.ServerSocket;
import flash.net.SharedObject;
import flash.net.Socket;
import flash.utils.ByteArray;
import flash.utils.Timer;

import mx.controls.Alert;

import spark.components.Button;
import spark.components.HGroup;
import spark.components.TextInput;

//EMBEDS
import flash.net.NetworkInterface;
import flash.net.InterfaceAddress;

[Embed (source="../lib/emb/snd/Update.mp3")]
private var _updateSoundClass:Class;

[Embed (source="../lib/emb/txt/root.html", mimeType="application/octet-stream")]
private var _rootHTMLClass:Class;

[Embed (source="../lib/emb/txt/manifest.plist", mimeType="application/octet-stream")]
private var _manifestPLISTClass:Class;

[Embed (source="../lib/emb/img/defaultIcon57.png", mimeType="application/octet-stream")]
private var _defaultIcon57Class:Class;

[Embed (source="../lib/emb/img/defaultIcon72.png", mimeType="application/octet-stream")]
private var _defaultIcon72Class:Class;

[Embed (source="../lib/emb/img/defaultIcon512.png", mimeType="application/octet-stream")]
private var _defaultIcon512Class:Class;

//VARS

private var _serverSocket:ServerSocket;

private var _mimeTypes:Object = new Object();

private var _serverIP:String;

private var _ipaFile:File;

private var _versionIncrementer:int = 0;

private var _fileCheckTimer:Timer;
private var _lastModifiedTime:Number;

private var _so:SharedObject;

private var _hasIPAChanged:Boolean = false;


//start 'er up
private function init():void
{
	_so = SharedObject.getLocal("EasyIPA")
		
	if(_so.data.versionIncrementer)
	{
		_versionIncrementer = int(_so.data.versionIncrementer);
	}
		
	_mimeTypes[".html"] = "text/html";
	_mimeTypes[".png"] 	= "image/png";
	_mimeTypes[".plist"] = "text/xml";
	_mimeTypes[".ipa"] 	= "application/octet-stream";
	
	_serverIP = getLocalIP();

	if(_serverIP == "")
	{
		_serverIP = "UNKNOWN_IP" //unknown ip
	}
	
	serverIPInput.text = _serverIP;

	_ipaFile = new File(_so.data.ipaPath ? _so.data.ipaPath : File.applicationDirectory.nativePath+"NotARealFile.txt"); 

	if(_ipaFile.exists) ipaFileInput.text = _ipaFile.nativePath;
	
	NativeApplication.nativeApplication.addEventListener(Event.EXITING, kill);
	
	setupTextInput(bundleIDInput);
	setupTextInput(appNameInput);
	setupTextInput(versionNumberInput);
	
	_lastModifiedTime = 0;	
	
	if(_ipaFile.exists)
	{
		_lastModifiedTime = _ipaFile.modificationDate.getTime();
		_hasIPAChanged = true;
	}
	
	_fileCheckTimer = new Timer(3000);
	_fileCheckTimer.addEventListener(TimerEvent.TIMER, onTimerTick);
	_fileCheckTimer.start();

}

private function getLocalIP():String
{
	var localIP:String = "UNKNOWN_IP";
		
	for each(var networkInterface:NetworkInterface in NetworkInfo.networkInfo.findInterfaces())
	{
		for each(var networkAddress:InterfaceAddress in networkInterface.addresses)
		{
			var address:String = networkAddress.address.toString();
			if(address.slice(0,3) == "192")
			{
				localIP = address;
			}
		}
	}
		
	return localIP;
}

//stop the server before the app exits
private function kill(e:Event):void
{
	stopServer();
}

//we use this to check if the ipa has been modified
private function onTimerTick(e:TimerEvent):void 
{
	if(_ipaFile.exists)
	{
		var modifiedTime:Number = _ipaFile.modificationDate.getTime();
		
		if(_lastModifiedTime != modifiedTime)
		{
			_lastModifiedTime = modifiedTime;
			log.text += "IPA file has been updated!\n";
			Sound(new _updateSoundClass()).play();
			_hasIPAChanged = true;
		}
	}
}

//toggle the server on and off
private function toggleServer():void
{
	if(_serverSocket)
	{
		stopServer();
	}
	else
	{
		startServer();
	}
}

//start the server
private function startServer():void
{
	try
	{
		if(_serverSocket) _serverSocket.close();
		
		_serverSocket = new ServerSocket();
		_serverSocket.addEventListener(Event.CONNECT, onSocketConnect);
		_serverSocket.bind(Number(portInput.text));
		_serverSocket.listen();
		log.text += "Listening on port " + portInput.text + "...\n";
		statusLabel.text = "Status: Running";
		statusAddress.text = "http://" + serverIPInput.text + ":" + Number(portInput.text).toString();
		toggleButton.label = "Stop Server";
	}
	catch (error:Error)
	{
		Alert.show("Port " + portInput.text + " may be in use. Enter another port number and try again.\n(" + error.message +")", "Error");
		statusLabel.text = "Status: Not started"
		statusAddress.text = "";
		toggleButton.label = "Start Server";
		_serverSocket = null;
	}
}

//stop the server
private function stopServer():void
{
	statusLabel.text = "Status: Stopped";
	statusAddress.text = "";
	if(_serverSocket) _serverSocket.close();
	_serverSocket = null;
	toggleButton.label = "Start Server";
	log.text += "Closing connection...\n\n"
}

//start listening for socket data once we have a connection
private function onSocketConnect(event:ServerSocketConnectEvent):void
{
	var socket:Socket = event.socket;
	socket.addEventListener(ProgressEvent.SOCKET_DATA, onSocketData);
}

//handle the socket data
private function onSocketData(event:ProgressEvent):void
{
	try
	{
		var socket:Socket = event.target as Socket;
		var bytes:ByteArray = new ByteArray();
		socket.readBytes(bytes);
		var request:String = "" + bytes;
		var filePath:String = request.substring(4, request.indexOf("HTTP/") - 1);
		
		log.text += "Requesting file at '"+filePath+"'\n";
		
		getFileForPath(filePath, socket);
		
		socket.flush();
		socket.close();
	}
	catch (error:Error)
	{
		Alert.show(error.message, "Error");
	}
}

//return the right file for the request
//most of the files have been embedded as bytes
//for text files, we do some basic token replacement before returning them
private function getFileForPath(filePath:String, socket:Socket):void
{
	var success:Boolean = true;
	var fileType:String = "";
	var contentBytes:ByteArray = new ByteArray();
	
	if(filePath == "/")
	{
		fileType = "html";
		success = readTextIntoBytes(_rootHTMLClass, contentBytes)
		_hasIPAChanged = false;
	}
	else if(filePath == "/manifest.plist")
	{
		fileType = "plist";
		_versionIncrementer++;
		_so.data.versionIncrementer = _versionIncrementer;
		_so.flush();
		success = readTextIntoBytes(_manifestPLISTClass, contentBytes)
	}
	else if(filePath == "/app.ipa")
	{
		fileType = "ipa";
		success = readFileIntoBytes(_ipaFile, contentBytes)
	}
	else if(filePath == "/iconNormal.png")
	{
		fileType = "png";
		var defaultIcon72Bytes:ByteArray = ByteArray(new _defaultIcon72Class());
		defaultIcon72Bytes.readBytes(contentBytes);
		success = true;
	}
	else if(filePath == "/icon512.png")
	{
		fileType = "png";
		var defaultIcon512Bytes:ByteArray = ByteArray(new _defaultIcon512Class());
		defaultIcon512Bytes.readBytes(contentBytes);
		success = true;
	}
	else 
	{
		success = false;
	}
	
	if(success)
	{
		log.text += "Returning the file, fileType was "+fileType+"\n"
		socket.writeUTFBytes("HTTP/1.1 200 OK\n");
		socket.writeUTFBytes("Content-Type: " + getMimeType(filePath) + "\n\n");
		socket.writeBytes(contentBytes);
	}
	else
	{
		log.text += "Returning 404, fileType was "+fileType+"\n"
		socket.writeUTFBytes("HTTP/1.1 404 Not Found\n");
		socket.writeUTFBytes("Content-Type: text/html\n\n");
		socket.writeUTFBytes("<html><body><h2>Page Not Found</h2></body></html>");
	}
}

//read the file (usually the ipa file) into the bytearray
private function readFileIntoBytes(file:File, bytes:ByteArray, defaultBytesClass:Class = null):Boolean
{
	if(!file.exists || file.isDirectory)
	{	
		if(defaultBytesClass)
		{
			var defaultBytes:ByteArray = new defaultBytesClass();
			defaultBytes.readBytes(bytes);
			log.text += "File was invalid. Using default instead\n"
			return true;
		}
		else
		{
			log.text += "File was invalid\n"
			return false;
		}
	}
	else
	{
		var stream:FileStream = new FileStream();
		stream.open( file, FileMode.READ );
		stream.readBytes(bytes);
		stream.close();
		return true;
	}
}

//read the text into the byte array, but first do some token replacement
private function readTextIntoBytes(baseTextFileClass:Class, bytes:ByteArray):Boolean
{
	var baseTextBytes:ByteArray = ByteArray(new baseTextFileClass());
	var text:String = baseTextBytes.readUTFBytes(baseTextBytes.bytesAvailable);
	
	var versionNumberString:String = versionNumberInput.text + "." + String(_versionIncrementer); //ensure the version is *always* increasing
	
	//do search replace
	text = replaceText(text, "{{BASE_URL}}", "http://"+ serverIPInput.text + ":" + Number(portInput.text).toString());
	text = replaceText(text, "{{APP_VERSION}}", versionNumberString);
	text = replaceText(text, "{{APP_TITLE}}", appNameInput.text);
	text = replaceText(text, "{{BUNDLE_ID}}", bundleIDInput.text);
	text = replaceText(text, "{{HAS_IPA_CHANGED}}", String(_hasIPAChanged));

	bytes.writeUTFBytes(text);

	return true;
}

//helper replacement function, a little overkill, but whatevs
private function replaceText(source:String, searcher:String,replacer:String):String
{
	return source.replace(new RegExp(searcher,"g"),replacer);
}

//return the correct mimetype for the file extension
private function getMimeType(path:String):String
{
	var mimeType:String;
	var index:int = path.lastIndexOf(".");
	if (index > -1)
	{
		mimeType = _mimeTypes[path.substring(index)];
	}
	return mimeType == null ? "text/html" : mimeType; // default to text/html for unknown mime types
}

//open the file browser to search for the ipa file
private function browseForFile(e:Event):void
{
	var group:HGroup = HGroup(Button(e.target).parent);
	var file:File = null;
	var message:String = "";
	
	if(group.id == "groupIPA")
	{
		file = _ipaFile;
		message = "Please select the IPA file...";
	}
	
	file.addEventListener(Event.SELECT, onFileSelect);
	file.browseForOpen(message);
}

//when an ipa file has been selected
private function onFileSelect(e:Event):void
{
	var file:File = File(e.currentTarget);
	if(file == _ipaFile) 
	{
		ipaFileInput.text = file.nativePath;
		_so.data.ipaPath = file.nativePath;
	}
	_so.flush();
}

//store the text in the shared object
private function onInputChange(e:Event):void 
{
	storeTextInput(spark.components.TextInput(e.target));
}

//store text in the sharedObject so it persists
private function storeTextInput(textInput:spark.components.TextInput):void
{
	if(_so && _so.data)
	{
		_so.data[textInput.id] = textInput.text;
		_so.flush();
	}
}

//load the text from the shared object
private function setupTextInput(textInput:spark.components.TextInput):void
{
	if(_so && _so.data && _so.data[textInput.id])
	{
		textInput.text = _so.data[textInput.id]
	}
}

//Browse through the IPA(which is really just a zip file)
//and grab some basic info from the Info.plist
private function extractDataFromIPA():void
{
	var success:Boolean = false;
	
	log.text += "Beginning extraction...\n"
		
	if(_ipaFile.exists)
	{
		var ipaBytes:ByteArray = new ByteArray();
		var stream:FileStream = new FileStream();
		stream.open( _ipaFile, FileMode.READ );
		stream.readBytes(ipaBytes);
		stream.close();
		
		var zip:FZip = new FZip();
		zip.loadBytes(ipaBytes);
		
		for(var z:int = 0; z<zip.getFileCount(); z++)
		{
			try 
			{
				var zipFile:FZipFile = zip.getFileAt(z);
				
				if(zipFile.filename.indexOf("Info.plist") != -1)
				{
					var infoXML:XML = new XML(zipFile.content.readUTFBytes(zipFile.content.bytesAvailable));
					
					var children:XMLList = infoXML.dict[0].children();
					
					for each(var node:XML in children)
					{
						if(node.name() == "key")
						{
							if(node.toString() == "CFBundleVersion")
							{
								log.text += "Found version number.\n"
								versionNumberInput.text = children[node.childIndex()+1].toString();
								storeTextInput(versionNumberInput);
							}
							else if(node.toString()== "CFBundleDisplayName")
							{
								log.text += "Found app name.\n"
								appNameInput.text = children[node.childIndex()+1].toString();
								storeTextInput(appNameInput);
							}
							else if(node.toString() == "CFBundleIdentifier")
							{
								log.text += "Found bundle identifier.\n"
								bundleIDInput.text = children[node.childIndex()+1].toString();
								storeTextInput(bundleIDInput);
							}
						}
					}
					
					break;
				}
				
				success = true;
			}
			catch (e:Error)
			{
				success = false;
				log.text += "Couldn't parse IPA file.\n"
			}
			
			
		}
	}
	else
	{
		log.text += "IPA file doesn't exist.\n"
		success = false;
	}
	
	if(success)
	{
		log.text += "Extraction complete.\n"
	}
	else
	{
		log.text += "Extraction failed!\n"
	}
	
}

//clear the log
private function clearLog():void
{
	log.text = "";
}









